@@ -1,6 +1,7 @@
 #![cfg_attr(test, feature(test))]
 
 use crate::buffer_proxy_iterator::BufferProxyIterator;
+use crate::feature_cell::FeatureCell;
 use crate::mtch::*;
 use crate::radix::*;
 use bitflags::bitflags;
@@ -103,7 +104,7 @@ pub struct Censor<I: Iterator<Item = char>> {
     replacements: u8,
     /// How many instances of censor replacement in the raw text?
     self_censoring: u8,
-    /// Is the input completely safe
+    /// Is the input completely safe.
     safe: bool,
     /// Where matches are kept after they are complete but may be cancelled due to false positives.
     pending_commit: Vec<Match>,
@@ -557,8 +558,7 @@ impl<I: Iterator<Item = char>> Iterator for Censor<I> {
 
             let pos = self.buffer.index();
 
-            let raw_c_upper = raw_c.is_uppercase();
-            self.uppercase = self.uppercase.saturating_add(raw_c_upper as u8);
+            self.uppercase = self.uppercase.saturating_add(raw_c.is_uppercase() as u8);
             let skippable = raw_c.is_punctuation() || raw_c.is_separator() || raw_c.is_other();
             let replacement = REPLACEMENTS.get(&raw_c);
 
@@ -622,16 +622,16 @@ impl<I: Iterator<Item = char>> Iterator for Censor<I> {
             let mut drain_start: Option<usize> = None;
             let mut safety_end = usize::MAX;
             let mut replacement_counted = false;
+            let raw_c_lower = raw_c.to_lowercase().next().unwrap();
 
             mem::swap(&mut self.matches, &mut self.matches_tmp);
             for c in replacement
                 .unwrap_or(&&*raw_c.encode_utf8(&mut [0; 4]))
                 .chars()
             {
-                let benign_replacement = raw_c_upper && c.is_lowercase();
+                let benign_replacement = c == raw_c_lower;
 
-                if !(c == raw_c
-                    || replacement_counted
+                if !(replacement_counted
                     || benign_replacement
                     || raw_c.is_ascii_alphabetic()
                     || (raw_c.is_ascii_digit()
@@ -652,9 +652,9 @@ impl<I: Iterator<Item = char>> Iterator for Censor<I> {
                             spaces: m.spaces.saturating_add(
                                 ((c == ' ' || c != raw_c) && self.separate && c != '\'') as u8,
                             ),
-                            replacements: m.replacements.saturating_add(
-                                (c != raw_c && !benign_replacement && !self.separate) as u8,
-                            ),
+                            replacements: m
+                                .replacements
+                                .saturating_add((!benign_replacement && !self.separate) as u8),
                             ..m
                         };
                         if let Some(existing) = self.matches.get(&undo_m) {
@@ -671,9 +671,9 @@ impl<I: Iterator<Item = char>> Iterator for Censor<I> {
                             spaces: m
                                 .spaces
                                 .saturating_add((c != raw_c && self.separate && c != '\'') as u8),
-                            replacements: m.replacements.saturating_add(
-                                (c != raw_c && !benign_replacement && !self.separate) as u8,
-                            ),
+                            replacements: m
+                                .replacements
+                                .saturating_add((!benign_replacement && !self.separate) as u8),
                             last: c,
                             ..m
                         };
@@ -838,6 +838,357 @@ impl<I: Iterator<Item = char> + Clone> CensorIter for I {
     }
 }
 
+#[cfg(feature = "context")]
+use std::collections::VecDeque;
+#[cfg(feature = "context")]
+use std::fmt::{Display, Formatter};
+#[cfg(feature = "context")]
+use std::time::{Duration, Instant};
+
+/// Context is useful for taking moderation actions on a per-user basis i.e. each user would get
+/// their own Context.
+#[cfg(feature = "context")]
+pub struct Context {
+    history: VecDeque<(String, Instant)>,
+    rate_limit: Duration,
+    burst: u8,
+    burst_used: u8,
+    suspicion: u8,
+    reports: u8,
+    total: u16,
+    total_inappropriate: u16,
+    muted_until: Option<Instant>,
+    only_safe_until: Option<Instant>,
+    rate_limited_until: Option<Instant>,
+    last_message: Option<Instant>,
+}
+
+#[cfg(feature = "context")]
+impl Context {
+    /// VecDeque internally adds 1 (resulting in 8), and then finds the next power of 2 (still 8).
+    const CAPACITY: usize = 7;
+
+    /// Track repetitions (spam) within this time frame.
+    const REPETITION_DURATION: Duration = Duration::from_secs(60);
+
+    /// How many repetitions are tolerated.
+    const REPETITION_LIMIT: usize = 3;
+
+    /// rate_limit is minimum time between messages.
+    /// burst allows a certain amount of messages beyond the rate limit.
+    pub fn new(rate_limit: Duration, burst: u8) -> Self {
+        Self {
+            history: VecDeque::with_capacity(Self::CAPACITY),
+            rate_limit,
+            burst,
+            burst_used: 0,
+            suspicion: 0,
+            reports: 0,
+            total: 0,
+            total_inappropriate: 0,
+            only_safe_until: None,
+            rate_limited_until: None,
+            muted_until: None,
+            last_message: None,
+        }
+    }
+
+    /// Returns None if expired is None or has been reached, resulting in expiry being set to None.
+    /// Otherwise, returns duration before expiry.
+    fn remaining_duration(expiry: &mut Option<Instant>, now: Instant) -> Option<Duration> {
+        if let Some(time) = *expiry {
+            if now >= time {
+                *expiry = None;
+                None
+            } else {
+                Some(time - now)
+            }
+        } else {
+            None
+        }
+    }
+
+    /// Takes user message, returns censored message trimmed of whitespace (if it should be sent)
+    /// or `BlockReason` (explaining why it should be blocked entirely).
+    pub fn process(&mut self, message: String) -> Result<String, BlockReason> {
+        let now = Instant::now();
+        let elapsed = self.last_message.map(|l| now - l).unwrap_or(Duration::ZERO);
+
+        let suspicion = self.suspicion.max(1).saturating_mul(self.reports.max(1));
+
+        // How convinced are we that the user is a bad actor.
+        let is_kinda_sus = suspicion >= 5;
+        let is_impostor = suspicion >= 25;
+
+        // Don't give bad actors the benefit of the doubt when it comes to meanness.
+        let meanness_threshold = if is_impostor {
+            Type::MILD_OR_HIGHER
+        } else if is_kinda_sus {
+            Type::MODERATE_OR_HIGHER
+        } else {
+            Type::SEVERE
+        };
+
+        let censor_threshold =
+            Type::PROFANE | Type::OFFENSIVE | Type::SEXUAL | (Type::MEAN & meanness_threshold);
+
+        // Don't give bad actors the benefit of letting their first character through.
+        let censor_first_character_threshold = if is_kinda_sus {
+            censor_threshold
+        } else {
+            // Mainly for protection against the n-word being discernible.
+            Type::OFFENSIVE & Type::SEVERE
+        };
+
+        let (mut censored, analysis) = Censor::from_str(trim_whitespace(&message))
+            .with_censor_threshold(censor_threshold)
+            .with_censor_first_character_threshold(censor_first_character_threshold)
+            .censor_and_analyze();
+
+        // Censoring can remove certain characters, rendering the prefix or suffix whitespace.
+        let trimmed_censored = trim_whitespace(&censored);
+        if trimmed_censored.len() < censored.len() {
+            censored = String::from(trimmed_censored);
+        }
+
+        self.total = self.total.saturating_add(1);
+        if analysis.is(Type::INAPPROPRIATE) {
+            self.total_inappropriate = self.total_inappropriate.saturating_add(1);
+        }
+
+        // Collecting suspicion.
+        let type_to_sus = |typ: Type| -> u8 {
+            let combined = analysis & typ;
+            if combined.is(Type::SEVERE) {
+                3
+            } else if combined.is(Type::MODERATE) {
+                2
+            } else if combined.is(Type::MILD) {
+                1
+            } else {
+                0
+            }
+        };
+
+        // Repetition detection.
+        self.history
+            .retain(|&(_, t)| now - t < Self::REPETITION_DURATION);
+        let mut recent_similar = 0;
+
+        for (recent_message, _) in &self.history {
+            if strsim::normalized_levenshtein(&recent_message, &message) >= 2.0 / 3.0 {
+                recent_similar += 1;
+            }
+        }
+
+        let mut new_suspicion = type_to_sus(Type::PROFANE | Type::OFFENSIVE | Type::SEXUAL)
+            .saturating_add(type_to_sus(Type::EVASIVE))
+            .saturating_add(type_to_sus(Type::SPAM));
+
+        if recent_similar >= 2 {
+            // Don't penalize as much for repeated messages, since an innocent user may repeat their
+            // message multiple times if it was erroneously detected.
+            new_suspicion /= 2;
+        }
+
+        if (is_kinda_sus && new_suspicion >= 4) || (is_impostor && new_suspicion >= 2) {
+            if let Some(only_safe_until) =
+                self.only_safe_until
+                    .unwrap_or(now)
+                    .checked_add(if self.reports > 0 {
+                        Duration::from_secs(10 * 60)
+                    } else {
+                        Duration::from_secs(5 * 60)
+                    })
+            {
+                self.only_safe_until = Some(only_safe_until);
+            }
+        }
+
+        self.suspicion = self.suspicion.saturating_add(new_suspicion);
+
+        let remaining_rate_limit = Self::remaining_duration(&mut self.rate_limited_until, now);
+
+        if let Some(dur) = Self::remaining_duration(&mut self.muted_until, now) {
+            Err(BlockReason::Muted(dur))
+        } else if censored.is_empty() {
+            Err(BlockReason::Empty)
+        } else if let Some(dur) = remaining_rate_limit.filter(|_| self.burst_used >= self.burst) {
+            Err(BlockReason::Spam(dur))
+        } else if recent_similar >= Self::REPETITION_LIMIT {
+            Err(BlockReason::Repetitious(recent_similar))
+        } else if analysis.is(Type::INAPPROPRIATE & Type::SEVERE) {
+            Err(BlockReason::Inappropriate(analysis))
+        } else if let Some(dur) = Self::remaining_duration(&mut self.only_safe_until, now)
+            .filter(|_| !analysis.is(Type::SAFE))
+        {
+            Err(BlockReason::Unsafe(dur))
+        } else {
+            if self.history.len() >= Self::CAPACITY {
+                self.history.pop_front();
+            }
+            self.history.push_back((message, now));
+            self.last_message = Some(now);
+            if self.rate_limit > Duration::ZERO {
+                self.burst_used = if remaining_rate_limit.is_some() {
+                    self.burst_used.saturating_add(1)
+                } else {
+                    self.burst_used.saturating_sub(
+                        (elapsed.as_nanos() / self.rate_limit.as_nanos()).min(u8::MAX as u128)
+                            as u8,
+                    )
+                };
+                if let Some(rate_limited_until) = self
+                    .rate_limited_until
+                    .unwrap_or(now)
+                    .checked_add(self.rate_limit)
+                {
+                    self.rate_limited_until = Some(rate_limited_until);
+                }
+            }
+            // Forgiveness.
+            self.suspicion = self
+                .suspicion
+                .saturating_sub((elapsed.as_secs() / 60).clamp(1, u8::MAX as u64) as u8);
+            Ok(censored)
+        }
+    }
+
+    /// Manually mute this user's messages for a duration. Overwrites any previous manual mute.
+    /// Passing `Duration::ZERO` will therefore un-mute.
+    pub fn mute_for(&mut self, duration: Duration) {
+        self.muted_until = Some(Instant::now() + duration);
+    }
+
+    /// Call if another user "reports" this user's message(s). The function of reports is for
+    /// suspicion of bad behavior to be confirmed faster.
+    pub fn report(&mut self) {
+        self.reports = self.reports.saturating_add(1);
+    }
+
+    /// Returns number of reports received via `Self::report()`. It is not guaranteed that the full
+    /// range of `usize` of reports will be counted (currently only `u8::MAX` are counted).
+    pub fn reports(&self) -> usize {
+        self.reports as usize
+    }
+
+    /// Returns total number of messages processed. It is not guaranteed that the full
+    /// range of `usize` of messages will be counted (currently only `u16::MAX` are counted).
+    pub fn total(&self) -> usize {
+        self.total as usize
+    }
+
+    /// Returns total number of messages processed that were `Type::INAPPROPRIATE`. It is not
+    /// guaranteed that the full range of `usize` of messages will be counted (currently only
+    /// `u16::MAX` are counted).
+    pub fn total_inappropriate(&self) -> usize {
+        self.total_inappropriate as usize
+    }
+}
+
+#[cfg(feature = "context")]
+impl Default for Context {
+    fn default() -> Self {
+        Self::new(Duration::from_secs(5), 3)
+    }
+}
+
+#[cfg(feature = "context")]
+#[derive(Copy, Clone, Debug, PartialEq)]
+#[non_exhaustive]
+pub enum BlockReason {
+    /// The particular message was *severely* inappropriate, more specifically, `Type`.
+    Inappropriate(Type),
+    /// Recent messages were generally inappropriate, and this message isn't on the safe list.
+    /// Try again after `Duration`.
+    Unsafe(Duration),
+    /// This message was too similar to `usize` recent messages.
+    Repetitious(usize),
+    /// Too many messages per unit time, try again after `Duration`.
+    Spam(Duration),
+    /// Manually muted for `Duration`.
+    Muted(Duration),
+    /// Message was, at least after censoring, completely empty.
+    Empty,
+}
+
+#[cfg(feature = "context")]
+impl BlockReason {
+    /// You may display `BlockReason` in any manner you choose, but this will return a reasonable
+    /// default warning to send to the user.
+    pub fn generic_str(self) -> &'static str {
+        match self {
+            Self::Inappropriate(_) => "Your message was held for severe profanity",
+            Self::Unsafe(_) => "You have been temporarily restricted due to profanity/spam",
+            Self::Repetitious(_) => "Your message was too similar to recent messages",
+            Self::Spam(_) => "You have been temporarily muted due to excessive frequency",
+            Self::Muted(_) => "You have been temporarily muted",
+            Self::Empty => "Your message was empty",
+        }
+    }
+
+    /// You may display `BlockReason` in any manner you choose, but this will return a reasonable
+    /// default warning to send to the user that includes some context (such as how long they are
+    /// muted for).
+    pub fn contextual_str(self) -> String {
+        match self {
+            Self::Inappropriate(typ) => String::from(if typ.is(Type::OFFENSIVE) {
+                "Your message was held for being highly offensive"
+            } else if typ.is(Type::SEXUAL) {
+                "Your message was held for being overly sexual"
+            } else if typ.is(Type::MEAN) {
+                "Your message was held for being overly mean"
+            } else {
+                "Your message was held for severe profanity"
+            }),
+            Self::Unsafe(dur) => format!(
+                "You have been restricted for {} due to profanity/spam",
+                FormattedDuration(dur)
+            ),
+            Self::Repetitious(count) => {
+                format!("Your message was too similar to {} recent messages", count)
+            }
+            Self::Spam(dur) => format!(
+                "You have been muted for {} due to excessive frequency",
+                FormattedDuration(dur)
+            ),
+            Self::Muted(dur) => format!("You have been muted for {}", FormattedDuration(dur)),
+            _ => String::from(self.generic_str()),
+        }
+    }
+}
+
+#[cfg(feature = "context")]
+struct FormattedDuration(Duration);
+
+#[cfg(feature = "context")]
+impl Display for FormattedDuration {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        if self.0 >= Duration::from_secs(3600) {
+            write!(f, "{}h", self.0.as_secs() / 3600)
+        } else if self.0 >= Duration::from_secs(60) {
+            write!(f, "{}m", self.0.as_secs() / 60)
+        } else {
+            write!(f, "{}s", self.0.as_secs().max(1))
+        }
+    }
+}
+
+/// Trims whitespace characters from both ends of a string, according to the definition of
+/// `crate::is_whitespace`.
+pub fn trim_whitespace(s: &str) -> &str {
+    s.trim_matches(is_whitespace)
+}
+
+/// Returns true iff the character is effectively whitespace. The definition of whitespace is broader
+/// than that of Unicode, because it includes control characters and a few additional blank characters.
+pub fn is_whitespace(c: char) -> bool {
+    // NOTE: The following characters are not detected by standard means but show up as blank.
+    // https://www.compart.com/en/unicode/U+2800
+    // https://www.compart.com/en/unicode/U+3164
+    c.is_whitespace() || c.is_other() || c == '\u{2800}' || c == '\u{3164}'
+}
+
 /// Adds a word, with the given type. The type can be `Type::SAFE`, or a combination of `Type::PROFANE`,
 /// `Type::Sexual`, `Type::Offensive`, `Type::Mean`, `Type::Mild`, `Type::Moderate`, and `Type::Severe`,
 /// but NOT both (can't be safe and unsafe).
@@ -869,7 +1220,7 @@ mod tests {
     use serial_test::serial;
     use std::fs::File;
     use std::io::BufReader;
-    use std::time::Instant;
+    use std::time::{Duration, Instant};
     use test::Bencher;
 
     #[test]
@@ -1212,6 +1563,113 @@ mod tests {
         assert!(test_safe.is(Type::SAFE));
     }
 
+    /*
+           Self::Inappropriate(_) => "Your message was held for severe profanity",
+           Self::Unsafe(_) => "You have been temporarily restricted due to profanity/spam",
+           Self::Repetitious(_) => "Your message was too similar to recent messages",
+           Self::Spam(_) => "You have been temporarily muted due to excessive frequency",
+           Self::Muted(_) => "You have been temporarily muted",
+           Self::Empty => "Your message was empty"
+    */
+
+    #[cfg(feature = "context")]
+    #[test]
+    #[serial]
+    fn context_inappropriate() {
+        use crate::{BlockReason, Context};
+
+        let mut ctx = Context::new(Duration::ZERO, 0);
+
+        assert_eq!(ctx.process(String::from("one")), Ok(String::from("one")));
+        assert!(matches!(
+            ctx.process(String::from("nigga")),
+            Err(BlockReason::Inappropriate(_))
+        ));
+    }
+
+    #[cfg(feature = "context")]
+    #[test]
+    #[serial]
+    fn context_unsafe() {
+        use crate::{BlockReason, Context};
+
+        let mut ctx = Context::new(Duration::ZERO, 0);
+
+        for _ in 0..30 {
+            ctx.report();
+        }
+
+        let res = ctx.process(String::from("shit"));
+        assert!(matches!(res, Err(BlockReason::Unsafe(_))), "1 {:?}", res);
+
+        let res = ctx.process(String::from("not common message"));
+        assert!(matches!(res, Err(BlockReason::Unsafe(_))), "2 {:?}", res);
+    }
+
+    #[cfg(feature = "context")]
+    #[test]
+    #[serial]
+    fn context_repetitious() {
+        use crate::{BlockReason, Context};
+
+        let mut ctx = Context::new(Duration::ZERO, 0);
+
+        for _ in 0..Context::REPETITION_LIMIT {
+            assert!(ctx.process(String::from("one")).is_ok());
+        }
+
+        let res = ctx.process(String::from("onne"));
+        assert!(matches!(res, Err(BlockReason::Repetitious(_))), "{:?}", res);
+    }
+
+    #[cfg(feature = "context")]
+    #[test]
+    #[serial]
+    fn context_spam() {
+        use crate::{BlockReason, Context};
+
+        let mut ctx = Context::new(Duration::from_millis(350), 2);
+
+        assert_eq!(ctx.process(String::from("one")), Ok(String::from("one")));
+        assert_eq!(ctx.process(String::from("two")), Ok(String::from("two")));
+        assert_eq!(
+            ctx.process(String::from("three")),
+            Ok(String::from("three"))
+        );
+        assert!(matches!(
+            ctx.process(String::from("four")),
+            Err(BlockReason::Spam(_))
+        ));
+
+        std::thread::sleep(Duration::from_secs(2));
+
+        assert_eq!(ctx.process(String::from("one")), Ok(String::from("one")));
+    }
+
+    #[cfg(feature = "context")]
+    #[test]
+    #[serial]
+    fn context_muted() {
+        use crate::{BlockReason, Context};
+
+        let mut ctx = Context::new(Duration::ZERO, 0);
+
+        ctx.mute_for(Duration::from_secs(5));
+
+        let res = ctx.process(String::from("hello"));
+        assert!(matches!(res, Err(BlockReason::Muted(_))), "{:?}", res);
+    }
+
+    #[cfg(feature = "context")]
+    #[test]
+    #[serial]
+    fn context_empty() {
+        use crate::{BlockReason, Context};
+
+        let mut ctx = Context::new(Duration::from_secs(1), 2);
+        assert_eq!(ctx.process(String::from("   ")), Err(BlockReason::Empty));
+    }
+
     #[allow(soft_unstable)]
     #[bench]
     fn bench_is_inappropriate(b: &mut Bencher) {
@@ -1231,6 +1689,5 @@ mod tests {
     }
 }
 
-use crate::feature_cell::FeatureCell;
 use doc_comment::doctest;
 doctest!("../README.md");
